const Scene = require('Scene');
const Audio = require('Audio');
const Reactive = require('Reactive');
const Animation = require('Animation');
const Diagnostics = require('Diagnostics');

const privates = (() => {
    const map = new WeakMap();
    return obj => {
        let props = map.get(obj);
        if (!props) {
            props = {};
            map.set(obj, props);
        }
        return props;
    };
})();

class SoundModule {
    constructor(speaker, controller) {
        this.speaker = speaker;
        this.controller = controller;

        privates(this).volume = speaker.volume.pinLastValue();
        privates(this).volumeMultiplier = Reactive.val(1.0);
    }

    /**
     * Playback volume for this AudioObject
     * @type {number | ScalarSignal}
     */
    get volume() {
        return privates(this).volume;
    }

    /**
     * Playback volume for this AudioObject
     * @param {ScalarSignal} vol
     */
    set volume(vol) {
        // Allow setting either a primitive or a ScalarSignal
        const volSignal = vol.pinLastValue ? vol : Reactive.val(vol);
        privates(this).volume = volSignal.pinLastValue();
        this.speaker.volume = volSignal.mul(privates(this).volumeMultiplier);
    }

    /**
     * Play the audio sample from the beginning of the file
     */
    play() {
        this.controller.reset();
        this.controller.setPlaying(true);
    }

    /**
     * Begin looping the audio sample from the beginning of the file
     */
    loop() {
        this.controller.reset();
        this.controller.setPlaying(true);
        this.controller.setLooping(true);
    }

    /**
     * Stop playing or looping the audio sample
     */
    stop() {
        this.controller.stop();
    }


    /**
     * @param {number} targetVolume         The target volume
     * @param {number} durationMilliseconds The fade time in milliseconds
     * @param {{():void}} [completion]      The completion callback function
     * @returns {Promise<void>}
     */
    fadeToVolume(targetVolume, durationMilliseconds, completion) {
        const driver = Animation.timeDriver({ durationMilliseconds: durationMilliseconds, loopCount: 1, mirror: false });
        const crossfadeA = Animation.animate(driver, Animation.samplers.linear(0, 1));
        const crossfadeB = Reactive.val(1.0).sub(crossfadeA);
        const signal = crossfadeA.mul(targetVolume).add(crossfadeB.mul(this.volume));
        this.volume = signal;
        driver.start();

        if (privates(this).existingFade) {
            privates(this).existingFade.stop();
        }

        privates(this).existingFade = driver;

        return new Promise(resolve => {
            const sub = driver.onCompleted().subscribe(() => {
                if (completion != undefined) {
                    completion();
                }
                resolve();
                sub.unsubscribe();
            })
        });
    };
}

/**
 * Generate an AudioObject from a connected Speaker and Audio Playback Controller
 * @param  {string} name            The name of the Audio Playback Controller and Speaker Object
 * @return {Promise<SoundModule>}   An Sound referencing the specified Speaker and Playback Controller
 */
export const Sound = async (name) => {
    const checkSpeakerNull = result => {
        if (result == undefined) {
            throw `Speaker not found: ${name}`;
        }
        return result;
    }

    try {
        const [speaker, controller] = await Promise.all([
            Scene.root.findFirst(name).then(checkSpeakerNull).catch(Diagnostics.log),
            Audio.getAudioPlaybackController(name).catch(Diagnostics.log)
        ]);

        return new SoundModule(speaker, controller);

    } catch (error) {
        throw error;
    }
}